<?php

declare(strict_types=1);

namespace Drupal\devel_visual\Controller;

use Drupal\Component\Utility\Html;
use Drupal\Core\Controller\ControllerBase;

use Drupal\Core\Entity\Display\EntityDisplayInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Url;
use Psr\Http\Message\ResponseInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

class PageEntityDisplayModes extends ControllerBase {

  /**
   * @var array<string, string>
   */
  protected array $chars = [
    'heavyCheckMark' => '&#9989;',
    'crossMark' => '&#10060;',
  ];

  /**
   * {@inheritdoc}
   *
   * @noinspection PhpMissingParentCallCommonInspection
   */
  public static function create(ContainerInterface $container): static {
    // @phpstan-ignore-next-line
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity_display.repository'),
    );
  }

  public function __construct(
    EntityTypeManager $entityTypeManager,
    protected EntityTypeBundleInfoInterface $entityTypeBundleInfo,
    protected EntityDisplayRepositoryInterface $entityDisplayRepository,
  ) {
    $this->entityTypeManager = $entityTypeManager;
  }

  /**
   * Page title callback.
   *
   * @see /devel_visual.routing.yml
   */
  public function pageTitle(string $type): string|\Stringable {
    return match ($type) {
      'view' => $this->t('Entity view modes'),
      'form' => $this->t('Entity form modes'),
      default => $this->t('Entity display modes: @type', ['@type' => $type]),
    };
  }

  /**
   * Page content callback.
   *
   * @phpstan-return int|array<string, mixed>|\Psr\Http\Message\ResponseInterface
   *
   * @see /devel_visual.routing.yml
   */
  public function pageBody(string $type): int|array|ResponseInterface {
    $build = [
      'content' => [],
    ];

    foreach ($this->getEntityTypes($type) as $entityType) {
      $build['content'][$entityType->id()] = $this->buildEntityType($type, $entityType);
    }

    return $build;
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function buildEntityType(string $type, EntityTypeInterface $entityType): array {
    return [
      '#type' => 'details',
      '#tree' => TRUE,
      '#title' => $entityType->id(),
      '#open' => TRUE,
      '#attributes' => [
        'id' => Html::getId("devel-visual-entity-display-modes--$type--{$entityType->id()}"),
        'class' => [
          'devel-visual-entity-display-modes',
          Html::getClass("devel-visual-entity-display-modes--$type"),
          Html::getClass("devel-visual-entity-display-modes--$type--{$entityType->id()}"),
        ],
      ],
      '#attached' => [
        'library' => [
          'devel_visual/entity_display_modes',
        ],
      ],

      'displayModes' => $this->buildEntityTypeDisplayModes($type, $entityType),
    ];
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function buildEntityTypeDisplayModes(string $type, EntityTypeInterface $entityType): array {
    $variables = $this->getVariables($type, $entityType);

    $build = [
      '#type' => 'table',
      '#colgroups' => [
        [
          [
            'class' => 'header',
          ],
        ],
      ],
      '#header' => [
        'corner' => $this->cellCornerHeader($variables),
        'default.renderer' => [
          'scope' => 'col',
          'data' => 'default',
        ],
      ],
      '#empty' => $this->t('No items'),
      '#rows' => [],
    ];

    foreach ($variables['displayModes'] as $displayModeShortId => $displayMode) {
      $build['#header']["$displayModeShortId.renderer"] = [
        'scope' => 'col',
        'data' => $displayModeShortId,
      ];
    }

    foreach ($variables['renderers'] as $bundleId => $displayModes) {
      $build['#rows'][$bundleId] = [
        'data' => array_fill_keys(array_keys($build['#header']), ['data' => '']),
      ];

      $build['#rows'][$bundleId]['data']['corner'] = $this->cellCornerBody($type, $entityType, $variables['bundles'][$bundleId]);
      foreach ($displayModes as $displayModeShortId => $renderer) {
        $build['#rows'][$bundleId]['data']["$displayModeShortId.renderer"] = $this->cellDisplayModeRenderer($renderer);
      }
    }

    return $build;
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function getVariables(string $type, EntityTypeInterface $entityType): array {
    $variables = [
      'type' => $type,
      'entityTypeId' => $entityType->id(),
      'operations' => $this->getOperations($type, $entityType),
      'bundles' => $this->getBundles($entityType),
      'displayModes' => $this->getDisplayModes($type, $entityType),
      'renderers' => [],
    ];

    /* @noinspection PhpUnhandledExceptionInspection */
    $edmStorage = $this
      ->entityTypeManager
      ->getStorage("entity_{$type}_display");

    foreach ($variables['bundles'] as $bundle) {
      $displayModeShortId = 'default';
      $displayId = sprintf('%s.%s.%s',
        $entityType->id(),
        $bundle['id'],
        $displayModeShortId,
      );
      // NOTE: Display "default" can be NULL when it is not saved yet.
      /** @var \Drupal\Core\Entity\Display\EntityDisplayInterface $display */
      $display = $edmStorage->load($displayId);

      $variables['renderers'][$bundle['id']] = [
        'default' => $this->getDisplayModeRenderer($displayModeShortId, $display),
      ];

      foreach ($variables['displayModes'] as $displayModeShortId => $displayMode) {
        $displayModeShortId = (string) $displayModeShortId;

        $displayId = sprintf('%s.%s.%s',
          $entityType->id(),
          $bundle['id'],
          $displayModeShortId,
        );

        // NOTE: Display NON "default" is NULL when it is not customized.
        /** @var \Drupal\Core\Entity\Display\EntityDisplayInterface $display */
        $display = $edmStorage->load($displayId);

        $variables['renderers'][$bundle['id']][$displayModeShortId] = $this->getDisplayModeRenderer($displayModeShortId, $display);
      }
    }

    return $variables;
  }

  /**
   * @phpstan-return array<string, array<string, mixed>>
   */
  protected function getOperations(string $type, EntityTypeInterface $entityType): array {
    $operations = [
      "$type.collection" => [
        'title' => $type === 'view' ? $this->t('View modes') : $this->t('Form modes'),
        'url' => Url::fromRoute(
          "entity.entity_{$type}_mode.collection",
          [],
          [
            'fragment' => sprintf(':~:text=%s', $entityType->getLabel()),
          ],
        ),
      ],
    ];

    $displayModeId = "entity_{$type}_mode";
    if ($this->entityTypeManager()->hasDefinition($displayModeId)) {
      /* @noinspection PhpUnhandledExceptionInspection */
      $displayMode = $this->entityTypeManager()->getDefinition($displayModeId);
      if ($displayMode->hasLinkTemplate('add-form')) {
        $routeName = "entity.entity_{$type}_mode.add_form";
        $routeParameters = [
          'entity_type_id' => $entityType->id(),
        ];
        $operations["$type.add_form"] = [
          'title' => $this->t('Add new'),
          'url' => Url::fromRoute($routeName, $routeParameters),
        ];
      }
    }

    return $operations;
  }

  /**
   * @return array<string, array<string, mixed>>
   */
  protected function getDisplayModes(string $type, EntityTypeInterface $entityType): array {
    return $type === 'view'
      ? $this->entityDisplayRepository->getViewModes($entityType->id())
      : $this->entityDisplayRepository->getFormModes($entityType->id());
  }

  /**
   * @return array<string, \Drupal\Core\Entity\EntityTypeInterface>
   */
  protected function getEntityTypes(string $type): array {
    $entityTypes = array_filter(
      $this->entityTypeManager()->getDefinitions(),
      $this->getEntityTypeFilter($type),
    );

    uasort($entityTypes, $this->getEntityTypeComparer());

    return $entityTypes;
  }

  protected function getEntityTypeFilter(string $type): callable {
    return function (EntityTypeInterface $entityType) use ($type): bool {
      if ($type === 'view' && $entityType->hasViewBuilderClass()) {
        return TRUE;
      }

      if ($type === 'form' && $entityType->hasFormClasses()) {
        return TRUE;
      }

      return FALSE;
    };
  }

  protected function getEntityTypeComparer(): callable {
    return function (EntityTypeInterface $a, EntityTypeInterface $b): int {
      return strnatcmp($a->id(), $b->id());
    };
  }

  /**
   * @phpstan-return array<string, array<string, mixed>>
   */
  protected function getBundles(EntityTypeInterface $entityType): array {
    $bundles = $this
      ->entityTypeBundleInfo
      ->getBundleInfo($entityType->id());
    foreach ($bundles as $bundleId => &$bundle) {
      $bundle['id'] = $bundleId;
    }
    uasort($bundles, $this->getBundleComparer());

    return $bundles;
  }

  protected function getBundleComparer(): callable {
    return function (array $a, array $b): int {
      return strnatcmp($a['id'], $b['id']);
    };
  }

  /**
   * @return array{handler: string, css: string, label: string|\Stringable, settings?: array<string, mixed>}
   */
  protected function getDisplayModeRenderer(string $displayModeShortId, ?EntityDisplayInterface $display): array {
    $handlerClassic = [
      'handler' => 'classic',
      'css' => 'classic',
      'label' => $this->t('Classic'),
    ];

    if (!$display) {
      return $displayModeShortId === 'default'
        ? $handlerClassic
        : ['handler' => 'inherit', 'css' => 'inherit', 'label' => $this->t('Inherit')];
    }

    $settings = $display->getThirdPartySettings('layout_builder');
    if (!empty($settings['enabled'])) {
      return [
        'handler' => 'layoutBuilder',
        'css' => sprintf(
          'lb-%s',
          empty($settings['allow_custom']) ? 'global' : 'custom',
        ),
        'label' => empty($settings['allow_custom']) ? $this->t('LB Global') : $this->t('LB Custom'),
        'settings' => $settings,
      ];
    }

    $settings = $display->getThirdPartySettings('ds');
    if (!empty($settings['layout']['id'])) {
      return [
        'handler' => 'displaySuite',
        'css' => 'display-suite',
        'label' => $this->t('DS @layout.id', ['@layout.id' => $settings['layout']['id']]),
        'settings' => $settings,
      ];
    }

    $settings = $display->getThirdPartySettings('field_layout');
    if (!empty($settings['id'])) {
      return [
        'handler' => 'fieldLayout',
        'css' => 'field-layout',
        'label' => $this->t('FL @layout.id', ['@layout.id' => $settings['id']]),
        'settings' => $settings,
      ];
    }

    return $handlerClassic;
  }

  /**
   * @phpstan-param array<string, mixed> $bundle
   */
  protected function getManageDisplayUrl(
    string $type,
    EntityTypeInterface $entityType,
    array $bundle,
    string $displayModeShortId,
  ): ?Url {
    $fieldUiBaseRoute = $entityType->get('field_ui_base_route');
    if (!$fieldUiBaseRoute) {
      return NULL;
    }

    $routeName = sprintf(
      'entity.entity_%s_display.%s.%s',
      $type,
      $entityType->id(),
      $displayModeShortId,
    );

    $parameters = [];
    $options = [];

    $bundleEntityTypeId = $entityType->getBundleEntityType();
    if ($bundleEntityTypeId) {
      $parameters[$bundleEntityTypeId] = $bundle['id'];
    }

    $url = NULL;
    try {
      $url = Url::fromRoute(
        $routeName,
        $parameters,
        $options,
      );
    }
    catch (\Throwable) {
      // Do nothing.
    }

    return $url;
  }

  /**
   * @phpstan-param array<string, mixed> $variables
   *
   * @phpstan-return array<string, mixed>
   */
  protected function cellCornerHeader(array $variables): array {
    return [
      'scope' => 'col',
      'data' => [
        '#type' => 'dropbutton',
        '#dropbutton_type' => 'small',
        '#links' => $variables['operations'],
      ],
    ];
  }

  /**
   * @phpstan-param array<string, mixed> $bundle
   *
   * @phpstan-return array<string, mixed>
   */
  protected function cellCornerBody(
    string $type,
    EntityTypeInterface $entityType,
    array $bundle,
  ): array {
    $manageDisplayUrl = $this->getManageDisplayUrl($type, $entityType, $bundle, 'default');
    if (!$manageDisplayUrl) {
      return [
        'header' => TRUE,
        'scope' => 'row',
        'data' => $bundle['id'],
      ];
    }

    return [
      'header' => TRUE,
      'scope' => 'row',
      'data' => [
        '#type' => 'link',
        '#title' => $bundle['id'],
        '#url' => $manageDisplayUrl,
      ],
    ];
  }

  /**
   * @phpstan-param array<string, mixed> $renderer
   *
   * @phpstan-return array<string, mixed>
   */
  protected function cellDisplayModeRenderer(array $renderer): array {
    return [
      'class' => [
        'display-mode-handler',
        "display-mode-handler--{$renderer['css']}",
      ],
      'data' => [
        '#markup' => $renderer['label'],
      ],
    ];
  }

}
