<?php

declare(strict_types=1);

namespace Drupal\devel_visual\FieldType;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityTypeRepositoryInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Url;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;
use Drupal\field\FieldConfigInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

abstract class ListBuilderBase extends ControllerBase {

  /**
   * @abstract
   */
  protected string $fieldTypeId = '';

  /**
   * @var array<string, mixed>
   */
  protected array $build = [];

  protected string $destination = '';

  /**
   * @var array<string, array<string, mixed>>
   */
  protected array $fieldTypes = [];

  /**
   * @var array<string, \Drupal\field\FieldConfigInterface>
   */
  protected array $fieldInstances = [];

  /**
   * @var array<string, \Drupal\field\FieldStorageConfigInterface>
   */
  protected array $fields = [];

  /**
   * @noinspection PhpMissingParentCallCommonInspection
   */
  public static function create(ContainerInterface $container): static {
    // @phpstan-ignore-next-line
    return new static(
      $container->get('entity_type.manager'),
      $container->get('entity_type.repository'),
      $container->get('plugin.manager.field.field_type'),
    );
  }

  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    protected EntityTypeRepositoryInterface $entityTypeRepository,
    protected FieldTypePluginManagerInterface $fieldTypeManager,
  ) {
    $this->entityTypeManager = $entityTypeManager;
  }

  /**
   * @phpstan-return array<string, mixed>
   *
   * @throws \Drupal\Core\Entity\Exception\AmbiguousBundleClassException
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Core\Entity\Exception\AmbiguousEntityClassException
   * @throws \Drupal\Core\Entity\Exception\NoCorrespondingEntityClassException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function content() {
    return $this
      ->initDestination()
      ->initFields()
      ->initBuild()
      ->contentAddBasicColumns()
      ->build;
  }

  protected function initDestination(): static {
    // @todo Get the current route path.
    $this->destination = "/admin/reports/devel_visual-field_type/{$this->fieldTypeId}";

    return $this;
  }

  protected function getFieldInstanceFilter(): callable {
    return function (FieldConfig $fieldInstance): bool {
      $expectedFieldTypeClass = $this->fieldTypes[$this->fieldTypeId]['class'];
      if (!array_key_exists($fieldInstance->getType(), $this->fieldTypes)) {
        // For example: "layout_section".
        return FALSE;
      }

      $fieldType = $this->fieldTypes[$fieldInstance->getType()];

      return $fieldType['class'] === $expectedFieldTypeClass
        || is_subclass_of($fieldType['class'], $expectedFieldTypeClass);
    };
  }

  /**
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\Exception\AmbiguousBundleClassException
   * @throws \Drupal\Core\Entity\Exception\AmbiguousEntityClassException
   * @throws \Drupal\Core\Entity\Exception\NoCorrespondingEntityClassException
   */
  protected function initFields(): static {
    $this->fieldTypes = $this->fieldTypeManager->getUiDefinitions();

    // @phpstan-ignore-next-line
    $this->fields = $this
      ->entityTypeManager()
      ->getStorage($this->entityTypeRepository->getEntityTypeFromClass(FieldStorageConfig::class))
      ->loadMultiple();

    /** @var array<string, \Drupal\field\Entity\FieldConfig> $fieldInstances */
    $fieldInstances = $this
      ->entityTypeManager()
      ->getStorage($this->entityTypeRepository->getEntityTypeFromClass(FieldConfig::class))
      ->loadMultiple();

    $this->fieldInstances = array_filter(
      $fieldInstances,
      $this->getFieldInstanceFilter(),
    );

    return $this;
  }

  protected function initBuild(): static {
    $this->build = [
      'content' => [
        '#type' => 'table',
        '#header' => [],
        '#empty' => $this->t('No items'),
        '#rows' => array_fill_keys(array_keys($this->fieldInstances), []),
      ],
    ];

    return $this;
  }

  protected function contentAddBasicColumns(): static {
    $this->build['content']['#header'] += [
      'type' => $this->t('Type'),
      'entity_type' => $this->t('Entity type'),
      'bundle' => $this->t('Bundle'),
      'field_name' => $this->t('Field name'),
      'cardinality' => $this->t('Cardinality'),
    ];

    $etm = $this->entityTypeManager();
    foreach ($this->fieldInstances as $instanceId => $fieldInstance) {
      $fieldId = $fieldInstance->getTargetEntityTypeId() . '.' . $fieldInstance->getName();
      $field = $this->fields[$fieldId];
      /* @noinspection PhpUnhandledExceptionInspection */
      $targetEntityType = $etm->getDefinition($fieldInstance->getTargetEntityTypeId());

      $editUrl = $this->fieldConfigEditUrl($targetEntityType, $fieldInstance);

      $this->build['content']['#rows'][$instanceId] += [
        'type' => [
          'data' => [
            '#markup' => $fieldInstance->getType(),
          ],
        ],
        'entity_type' => [
          'data' => [
            '#markup' => $targetEntityType->id(),
          ],
        ],
        'bundle' => [
          'data' => $editUrl ?
            [
              '#type' => 'link',
              '#title' => $fieldInstance->getTargetBundle(),
              '#url' => $editUrl,
            ]
            : [
              '#markup' => $fieldInstance->getTargetBundle(),
            ],
        ],
        'field_name' => [
          'data' => [
            '#markup' => $fieldInstance->getName(),
          ],
        ],
        'cardinality' => [
          'data' => $this->cellDataCardinality((int) $field->getCardinality()),
        ],
      ];
    }

    return $this;
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function cellDataBoolean(bool $value): array {
    return [
      '#markup' => $value ? '&#9989;' : '&#10060;',
    ];
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function cellDataBooleanStatus(bool $enabled, bool $value): array {
    return $enabled ?
      $this->cellDataBoolean($value)
      : [
        '#markup' => '',
      ];
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function cellDataCardinality(int $value): array {
    return [
      '#markup' => $value === -1 ? '&infin;' : $value,
    ];
  }

  protected function fieldConfigEditUrl(
    EntityTypeInterface $targetEntityType,
    FieldConfigInterface $fieldInstance,
  ): ?Url {
    try {
      return Url::fromRoute(
        "entity.field_config.{$targetEntityType->id()}_field_edit_form",
        [
          $targetEntityType->getBundleEntityType() => $fieldInstance->getTargetBundle(),
          'field_config' => $fieldInstance->id(),
        ],
        [
          'query' => [
            'destination' => $this->destination,
          ],
          'attributes' => [
            'title' => $this->t('Edit field instance settings'),
          ],
        ],
      );
    }
    catch (\Exception) {
      return NULL;
    }
  }

}
