<?php

declare(strict_types=1);

namespace Drupal\devel_visual\GraphProvider;

use Drupal\devel_visual\ComposerFinder;
use Fhaculty\Graph\Edge\Directed;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

class ComposerPackageDependencies extends GraphProviderBase {

  public const EVENT_ALTER = 'devel_visual.graph_provider.composer_package_dependencies.alter';

  protected string $graphName = 'composer_package_dependencies';

  /**
   * @var array<string, mixed>
   */
  protected array $json = [];

  /**
   * @var array<string, mixed>
   */
  protected array $lock = [];

  /**
   * {@inheritdoc}
   *
   * @noinspection PhpMissingParentCallCommonInspection
   */
  public static function create(ContainerInterface $container): static {
    // @phpstan-ignore-next-line
    return new static(
      $container->get('event_dispatcher'),
      $container->get('devel_visual.composer_finder'),
    );
  }

  public function __construct(
    protected EventDispatcherInterface $eventDispatcher,
    protected ComposerFinder $composerFinder,
  ) {
    parent::__construct($eventDispatcher);
  }

  protected function buildGraph(): static {
    $this
      ->initSource()
      ->addPackages($this->lock['packages'] ?? [], 'prod')
      ->addPackages($this->lock['packages-dev'] ?? [], 'prod')
      ->addJson($this->json['require'], 'prod')
      ->addJson($this->json['require-dev'], 'dev');

    return $this;
  }

  protected function initSource(): static {
    $this->json = [];
    $this->lock = [];

    $jsonFileName = $this->composerFinder->getComposerJson(DRUPAL_ROOT);
    if ($jsonFileName) {
      $this->json = (array) json_decode(file_get_contents($jsonFileName) ?: '{}', TRUE);
    }

    $this->json += [
      'name' => $this->getPackageNameFromDirectory(DRUPAL_ROOT),
      'require' => [],
      'require-dev' => [],
    ];

    $lockFileName = preg_replace('/\.json$/', '.lock', $jsonFileName);
    if ($lockFileName) {
      $this->lock = (array) json_decode(file_get_contents($lockFileName) ?: '{}', TRUE);
    }

    $this->lock += [
      'packages' => [],
      'packages-dev' => [],
    ];

    return $this;
  }

  /**
   * @phpstan-param array<array<string, mixed>> $packages
   */
  protected function addPackages(array $packages, string $relationship): static {
    foreach ($packages as $from) {
      $fromVertex = $this->graph->createVertex($from['name'], TRUE);
      $fromVertex->setAttribute('label', $from['name']);
      $fromVertex->setAttribute('type', $from['type'] ?? 'library');
      $fromVertex->setAttribute('version', ltrim($from['version'], 'v'));

      if (empty($from['require'])) {
        continue;
      }

      foreach ($from['require'] as $toName => $toVersion) {
        $toVertex = $this->graph->createVertex($toName, TRUE);

        $edge = new Directed($fromVertex, $toVertex);
        $edge->setAttribute('id', "{$from['name']}-{$toName}");
        $edge->setAttribute('version', $toVersion);
        $edge->setAttribute('relationship', $relationship);
      }
    }

    return $this;
  }

  /**
   * @phpstan-param array<string, string> $packages
   */
  protected function addJson(array $packages, string $relationship): static {
    $fromVertex = $this->graph->createVertex($this->json['name'], TRUE);
    foreach ($packages as $toName => $edgeVersion) {
      // @phpstan-ignore-next-line
      $toVertex = $this->graph->createVertex($toName, TRUE);

      $edge = new Directed($fromVertex, $toVertex);
      $edge->setAttribute('id', "{$this->json['name']}-{$toName}-$relationship");
      $edge->setAttribute('version', $edgeVersion);
      $edge->setAttribute('relationship', $relationship);
    }

    return $this;
  }

  protected function getPackageNameFromDirectory(string $absoluteDir): string {
    $parts = explode(DIRECTORY_SEPARATOR, $absoluteDir);

    return implode('/', array_slice($parts, -2));
  }

}
