<?php

declare(strict_types=1);

namespace Drupal\devel_visual\GraphProvider;

use Drupal\Core\Asset\LibraryDiscoveryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Extension\ThemeHandlerInterface;
use Fhaculty\Graph\Edge\Directed;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

class LibraryDependencies extends GraphProviderBase {

  public const EVENT_ALTER = 'devel_visual.graph_provider.library_dependencies.alter';

  protected string $graphName = 'library_dependencies';

  /**
   * {@inheritdoc}
   *
   * @noinspection PhpMissingParentCallCommonInspection
   */
  public static function create(ContainerInterface $container): static {
    // @phpstan-ignore-next-line
    return new static(
      $container->get('event_dispatcher'),
      $container->get('library.discovery'),
      $container->get('module_handler'),
      $container->get('theme_handler'),
    );
  }

  public function __construct(
    protected EventDispatcherInterface $eventDispatcher,
    protected LibraryDiscoveryInterface $libraryDiscovery,
    protected ModuleHandlerInterface $moduleHandler,
    protected ThemeHandlerInterface $themeHandler,
  ) {
    parent::__construct($eventDispatcher);
  }

  protected function buildGraph(): static {
    $libraries = $this->getLibraries();
    foreach ($libraries as $fromId => $library) {
      // @phpstan-ignore-next-line
      $fromVertex = $this->graph->createVertex($fromId, TRUE);
      $fromVertex->setAttribute('label', $fromId);
      $fromVertex->setAttribute('missing', FALSE);
      $fromVertex->setAttribute('sourceType', $library['sourceType']);

      if (empty($library['dependencies'])) {
        continue;
      }

      foreach ($library['dependencies'] as $toId) {
        $toVertex = $this->graph->createVertex($toId, TRUE);
        $toVertex->setAttribute('label', $toId);
        $toVertex->setAttribute('missing', !isset($libraries[$toId]));
        $toVertex->setAttribute('sourceType', 'unknown');

        $edge = new Directed($fromVertex, $toVertex);
        $edge->setAttribute('id', "{$fromId}-{$toId}");
      }
    }

    return $this;
  }

  /**
   * @phpstan-return array<string, array<string, mixed>>
   */
  protected function getLibraries(): array {
    $libraries = [];
    $sources = [
      'core' => ['core'],
      'module' => array_keys($this->moduleHandler->getModuleList()),
      'theme' => array_keys($this->themeHandler->listInfo()),
    ];
    foreach ($sources as $sourceType => $extensionNames) {
      foreach ($extensionNames as $extensionName) {
        $libs = $this->libraryDiscovery->getLibrariesByExtension($extensionName);
        foreach ($libs as $libName => $lib) {
          $libId = "$extensionName/$libName";
          $libraries[$libId] = $lib + [
            'id' => $libId,
            'sourceType' => $sourceType,
          ];
        }
      }
    }

    return $libraries;
  }

}
