<?php

declare(strict_types=1);

namespace Drupal\devel_visual\GraphProvider;

use Drupal\Core\Extension\Extension;
use Drupal\Core\Extension\ModuleExtensionList;
use Fhaculty\Graph\Edge\Directed;
use Fhaculty\Graph\Vertex;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;

class ModuleDependencies extends GraphProviderBase {

  public const EVENT_ALTER = 'devel_visual.graph_provider.module_dependencies.alter';

  protected string $graphName = 'module_dependencies';

  /**
   * @var \Drupal\Core\Extension\Extension[]
   */
  protected array $modules = [];

  /**
   * {@inheritdoc}
   *
   * @noinspection PhpMissingParentCallCommonInspection
   */
  public static function create(ContainerInterface $container): static {
    // @phpstan-ignore-next-line
    return new static(
      $container->get('event_dispatcher'),
      $container->get('extension.list.module'),
    );
  }

  public function __construct(
    protected EventDispatcherInterface $eventDispatcher,
    protected ModuleExtensionList $moduleExtensionList,
  ) {
    parent::__construct($eventDispatcher);
  }

  protected function buildGraph(): static {
    $this->modules = $this->moduleExtensionList->reset()->getList();
    foreach ($this->modules as $fromName => $module) {
      $fromVertex = $this->graph->createVertex($fromName, TRUE);
      $this->setVertexAttributes($fromVertex);

      if (empty($module->info['dependencies'])) {
        continue;
      }

      foreach ($this->parseModuleNamesFromDependencies($module->info['dependencies']) as $toName) {
        // @phpstan-ignore-next-line
        $toVertex = $this->graph->createVertex($toName, TRUE);
        $this->setVertexAttributes($toVertex);

        $edge = new Directed($fromVertex, $toVertex);
        $edge->setAttribute('id', "{$fromName}-{$toName}");
      }
    }

    return $this;
  }

  protected function setVertexAttributes(Vertex $vertex): static {
    $module = $this->modules[$vertex->getId()] ?? NULL;
    $this
      ->setVertexAttributeGroup($vertex, $module)
      ->setVertexAttributePackage($vertex, $module)
      ->setVertexAttributeProject($vertex, $module)
      ->setVertexAttributeType($vertex, $module)
      ->setVertexAttributeVersion($vertex, $module)
      ->setVertexAttributeInstalled($vertex, $module)
      ->setVertexAttributeExperimental($vertex, $module);

    return $this;
  }

  protected function setVertexAttributeGroup(Vertex $vertex, ?Extension $extension): static {
    if (!$extension) {
      return $this;
    }

    $attributeName = 'graphviz.group';

    $path = $extension->getPath();
    if (mb_strpos($path, 'core/modules/') === 0) {
      $vertex->setAttribute($attributeName, 'core');
    }
    elseif (mb_strpos($path, 'modules/contrib/') === 0) {
      $vertex->setAttribute($attributeName, 'contrib');
    }
    elseif (mb_strpos($path, 'modules/custom/') === 0) {
      $vertex->setAttribute($attributeName, 'custom');
    }
    else {
      $vertex->setAttribute($attributeName, 'other');
    }

    return $this;
  }

  protected function setVertexAttributePackage(Vertex $vertex, ?Extension $extension): static {
    if ($extension && !empty($extension->info['package'])) {
      $vertex->setAttribute('graphviz.package', $extension->info['package']);
    }

    return $this;
  }

  protected function setVertexAttributeProject(Vertex $vertex, ?Extension $extension): static {
    if ($extension && !empty($extension->info['project'])) {
      $vertex->setAttribute('graphviz.project', $extension->info['project']);
    }

    return $this;
  }

  protected function setVertexAttributeType(Vertex $vertex, ?Extension $extension): static {
    if ($extension) {
      $vertex->setAttribute('graphviz.type', $extension->getType());
    }

    return $this;
  }

  protected function setVertexAttributeVersion(Vertex $vertex, ?Extension $extension): static {
    if ($extension && !empty($extension->info['version'])) {
      $vertex->setAttribute('graphviz.version', $extension->info['version']);
    }

    return $this;
  }

  protected function setVertexAttributeInstalled(Vertex $vertex, ?Extension $extension): static {
    $attributeName = 'graphviz.installed';

    if (!$extension) {
      $vertex->setAttribute($attributeName, 0);

      return $this;
    }

    // @todo Check that the extension is installed or not.
    $vertex->setAttribute($attributeName, 1);

    return $this;
  }

  protected function setVertexAttributeExperimental(Vertex $vertex, ?Extension $extension): static {
    $vertex->setAttribute('graphviz.experimental', (int) !empty($extension->info['experimental']));

    return $this;
  }

  /**
   * @phpstan-param array<string> $dependencies
   *
   * @phpstan-return array<string>
   */
  protected function parseModuleNamesFromDependencies(array $dependencies): array {
    $moduleNames = [];

    foreach ($dependencies as $dependency) {
      $moduleNames[] = $this->parseModuleNameFromDependency($dependency);
    }

    return $moduleNames;
  }

  protected function parseModuleNameFromDependency(string $dependency): string {
    $pattern = '/^((?P<vendor>[^:]+):){0,1}(?P<name>[^\s]+)(\s+(?P<version>.+)){0,1}$/';

    $matches = ['name' => ''];
    preg_match($pattern, $dependency, $matches);

    return $matches['name'] ?? '';
  }

}
