<?php

declare(strict_types=1);

namespace Drupal\devel_visual\ImageStyle;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\image\ImageStyleListBuilder;
use Drupal\image\ImageStyleStorageInterface;
use Sweetchuck\Utils\Comparer\ComparerInterface;
use Sweetchuck\Utils\Comparer\PredefinedComparer;
use Symfony\Component\DependencyInjection\ContainerInterface;

class ListBuilder extends ImageStyleListBuilder {

  /**
   * {@inheritdoc}
   *
   * @noinspection PhpMissingParentCallCommonInspection
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type): static {
    $etm = $container->get('entity_type.manager');

    // @phpstan-ignore-next-line
    return new static(
      $entity_type,
      $etm->getStorage($entity_type->id()),
      $container->get('file_url_generator'),
      $container->get('config.factory'),
    );
  }

  public function __construct(
    EntityTypeInterface $entity_type,
    ImageStyleStorageInterface $storage,
    protected FileUrlGeneratorInterface $fileUrlGenerator,
    protected ConfigFactoryInterface $configFactory,
  ) {
    parent::__construct($entity_type, $storage);
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-return array<string, mixed>
   */
  public function render() {
    $columnComparer = $this->getColumnComparer();

    $build = parent::render();

    $build['table']['#header']['id'] = $this->t('ID');
    uksort($build['table']['#header'], $columnComparer);

    /** @var array<string> $imageStyleIds */
    $imageStyleIds = array_keys($build['table']['#rows']);
    if (!$imageStyleIds) {
      return $build;
    }

    /** @var \Drupal\image\ImageStyleInterface[] $imageStyles */
    $imageStyles = $this->storage->loadMultiple($imageStyleIds);
    foreach ($build['table']['#rows'] as $imageStyleId => &$row) {
      $row['id'] = [
        'header' => TRUE,
        'data' => [
          '#type' => 'html_tag',
          '#tag' => 'strong',
          '#value' => $imageStyles[$imageStyleId]->id(),
        ],
      ];
      $row['label'] = [
        'header' => TRUE,
        'data' => [
          '#type' => 'html_tag',
          '#tag' => 'strong',
          '#value' => $imageStyles[$imageStyleId]->label(),
        ],
      ];
      $row['operations']['header'] = TRUE;
      uksort($row, $columnComparer);
    }

    $originalPath = $this
      ->configFactory
      ->get('image.settings')
      ->get('preview_image');
    foreach ($imageStyles as $imageStyleId => $imageStyle) {
      $details = [
        'data' => [
          'effects' => [
            'colspan' => 1,
            'data' => [],
          ],
          'preview' => [
            'colspan' => 2,
            'data' => [
              '#theme' => 'image_style',
              '#uri' => $originalPath,
              '#alt' => $this->t('Sample modified image'),
              '#title' => '',
              '#style_name' => $imageStyle->id(),
            ],
          ],
        ],
      ];
      foreach ($imageStyle->getEffects() as $effect) {
        $details['data']['effects']['data'][$effect->getUuid() . '.label'] = [
          '#type' => 'html_tag',
          '#tag' => 'span',
          '#value' => $effect->label(),
        ];
        $details['data']['effects']['data'][$effect->getUuid() . '.summary'] = $effect->getSummary();
        $details['data']['effects']['data'][$effect->getUuid() . '.separator'] = [
          '#markup' => '<br />',
        ];
      }
      $build['table']['#rows']["$imageStyleId.details"] = $details;
    }

    uksort($build['table']['#rows'], $this->getRowComparer($imageStyleIds));

    return $build;
  }

  /**
   * @phpstan-return \Sweetchuck\Utils\Comparer\ComparerInterface<string>
   */
  protected function getColumnComparer(): ComparerInterface {
    $comparer = new PredefinedComparer();
    $comparer->setOptions([
      'weights' => [
        'id' => 0,
        'label' => 1,
        'operations' => 99,
      ],
    ]);

    return $comparer;
  }

  /**
   * @phpstan-param array<string> $entityIds
   *
   * @phpstan-return callable(string, string)
   */
  protected function getRowComparer(array $entityIds): callable {
    $weights = array_flip($entityIds);
    foreach ($weights as $entityId => $weight) {
      $weights["$entityId.details"] = $weight + 0.1;
    }

    return function (string $a, string $b) use ($weights) {
      return $weights[$a] <=> $weights[$b];
    };
  }

}
