<?php

declare(strict_types=1);

namespace Drupal\devel_visual\NodeType;

use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\node\NodeTypeInterface;
use Drupal\node\NodeTypeListBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;

class ListBuilder extends NodeTypeListBuilder {

  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * @var array<string, string>
   */
  protected array $chars = [
    'heavyCheckMark' => '&#9989;',
    'crossMark' => '&#10060;',
  ];

  /**
   * {@inheritdoc}
   */
  public static function createInstance(ContainerInterface $container, EntityTypeInterface $entity_type) {
    $etm = $container->get('entity_type.manager');

    /* @noinspection PhpUnhandledExceptionInspection */
    // @phpstan-ignore-next-line
    return new static(
      $entity_type,
      $etm->getStorage($entity_type->id()),
      $etm,
    );
  }

  /**
   * {@inheritdoc}
   */
  public function __construct(
    EntityTypeInterface $entity_type,
    EntityStorageInterface $storage,
    EntityTypeManagerInterface $entityTypeManager,
  ) {
    $this->entityTypeManager = $entityTypeManager;
    parent::__construct($entity_type, $storage);
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-return array<string, \Drupal\node\NodeTypeInterface>
   */
  public function load() {
    $entity_ids = $this->getEntityIds();
    /** @var array<string, \Drupal\node\NodeTypeInterface> $entities */
    $entities = $this->storage->loadMultipleOverrideFree($entity_ids);

    uasort(
      $entities,
      function (NodeTypeInterface $a, NodeTypeInterface $b): int {
        return strnatcasecmp((string) $a->id(), (string) $b->id());
      },
    );

    return $entities;
  }

  /**
   * {@inheritdoc}
   *
   * @phpstan-return array<string, mixed>
   */
  public function buildHeader(): array {
    return [
      'id' => $this->t('ID'),
      'preview_mode' => $this->t('Preview mode'),
      'display_submitted' => $this->t('Display submitted'),
      'new_revision' => $this->t('New revision'),
      'status' => $this->t('Published'),
      'promote' => $this->t('Promote'),
      'sticky' => $this->t('Sticky'),
      'menu_options' => $this->t('Available menus'),
      'menu_parent' => $this->t('Default parent link'),
      'operations' => $this->t('Operations'),
    ];
  }

  /**
   * @param \Drupal\node\NodeTypeInterface $entity
   *
   * @phpstan-return array<string, mixed>
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function buildRow(EntityInterface $entity): array {
    $values = $this->extractDefaultValues($entity);

    return [
      'id' => [
        'data' => [
          '#type' => 'link',
          '#title' => $entity->id(),
          '#url' => $this->ensureDestination($entity->toUrl('edit-form')),
        ],
      ],
      'preview_mode' => [
        'data' => $this->cellDataPreview($values['preview_mode']),
      ],
      'display_submitted' => [
        'data' => $this->cellDataBoolean((bool) $values['display_submitted']),
      ],
      'new_revision' => [
        'data' => $this->cellDataBoolean((bool) $values['new_revision']),
      ],
      'status' => [
        'data' => $this->cellDataBoolean((bool) $values['status']),
      ],
      'promote' => [
        'data' => $this->cellDataBoolean((bool) $values['promote']),
      ],
      'sticky' => [
        'data' => $this->cellDataBoolean((bool) $values['sticky']),
      ],
      'menu_options' => [
        'data' => implode(
          ', ',
          $entity->getThirdPartySetting('menu_ui', 'available_menus', []),
        ),
      ],
      'menu_parent' => [
        'data' => $entity->getThirdPartySetting('menu_ui', 'parent', ''),
      ],
      'operations' => [
        'data' => $this->buildOperations($entity),
      ],
    ];
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function extractDefaultValues(NodeTypeInterface $bundle): array {
    /* @noinspection PhpUnhandledExceptionInspection */
    /** @var \Drupal\node\NodeInterface $node */
    $node = $this
      ->entityTypeManager
      ->getStorage('node')
      ->create([
        'type' => $bundle->id(),
      ]);

    return [
      'id' => $bundle->id(),
      'preview_mode' => $bundle->getPreviewMode(),
      'display_submitted' => $bundle->displaySubmitted() ? '1' : '0',
      'new_revision' => $bundle->shouldCreateNewRevision() ? '1' : '0',
      'status' => $node->get('status')->getValue()[0]['value'],
      'promote' => $node->get('promote')->getValue()[0]['value'],
      'sticky' => $node->get('sticky')->getValue()[0]['value'],
      'menu_options' => implode(
        ', ',
        $bundle->getThirdPartySetting('menu_ui', 'available_menus', []),
      ),
      'menu_parent' => $bundle->getThirdPartySetting('menu_ui', 'parent', ''),
    ];
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function cellDataBoolean(bool $value): array {
    return [
      '#markup' => $value ? $this->chars['heavyCheckMark'] : $this->chars['crossMark'],
    ];
  }

  /**
   * @phpstan-return array<string, mixed>
   */
  protected function cellDataPreview(int $value): array {
    return [
      '#markup' => match ($value) {
        0 => $this->t('Disabled'),
        1 => $this->t('Optional'),
        2 => $this->t('Required'),
        default => $value,
      },
    ];
  }

}
